import torch
from torch_geometric.data import Batch


def batch_to_Exyz(batch: Batch) -> Batch:
    zalphar: torch.Tensor = batch.x[..., [1, 2, 3]]
    num_alpha = 16
    # num_z = 45
    # num_r = 9
    z, alpha, r = zalphar.T
    # shift idx by one to go from (min,max)
    # (0, 1-1/num) -> (1/num,1)
    # z = (z + 1) / num_z
    # r = (r + 1) / num_r
    alpha = (alpha + 1) / num_alpha * torch.pi * 2

    y = r * torch.cos(alpha)
    x = r * torch.sin(alpha)

    batch.xyz = torch.stack([x, y, z]).T

    #       r
    #       ^
    #     / |
    #    /  |
    #   /   |
    # / θ)  |
    # ------> z

    theta = torch.arctan(r / z)
    if (theta.abs() > torch.pi).any():
        raise RuntimeError("θ not in forward direction")
    if (theta < 0).any():
        raise RuntimeError("θ not in forward direction")
    batch.eta = -torch.log(torch.tan(theta / 2.0)).reshape(-1, 1)
    batch.phi = alpha.reshape(-1, 1)

    return batch



import torch
from torch_geometric.data import Batch

from fgsim.config import conf

from .convcoord import batch_to_Exyz
from .pca import fpc_from_batch
from .shower import analyze_layers, cyratio, response, sphereratio
from .voxelize import sum_dublicate_hits

alphapos = conf.loader.x_features.index("alpha")
num_alpha = 16


def postprocess(batch: Batch, sim_or_gen: str) -> Batch:
    if sim_or_gen == "gen":
        alphas = batch.x[..., alphapos].clone()

        shift = torch.randint(0, num_alpha, (batch.batch[-1] + 1,)).to(
            alphas.device
        )[batch.batch]
        alphas = alphas.clone() + shift.float()
        alphas[alphas > num_alpha - 1] -= num_alpha

        batch.x[..., alphapos] = alphas

    batch = sum_dublicate_hits(batch, forbid_dublicates=sim_or_gen == "sim")
    batch = batch_to_Exyz(batch)
    metrics: list[str] = conf.training.val.metrics
    if "sphereratio" in metrics:
        batch["sphereratio"] = sphereratio(batch)
    if "cyratio" in metrics:
        batch["cyratio"] = cyratio(batch)
    if "fpc" in metrics:
        batch["fpc"] = fpc_from_batch(batch)
    if "showershape" in metrics:
        batch["showershape"] = analyze_layers(batch)
    if "response" in metrics:
        batch["response"] = response(batch)

    return batch



from typing import Dict

from matplotlib.figure import Figure

from fgsim.config import conf
from fgsim.plot.infolut import var_to_label

from .ratioplot import ratioplot


def hist1d(
    sim, gen, ftxname: str, bins=None, energy_weighted=False
) -> Dict[str, Figure]:
    epos = conf.loader.x_ftx_energy_pos
    ename = conf.loader.x_features[epos]

    sim_features = _exftxt(sim[ftxname])
    gen_features = _exftxt(gen[ftxname])
    fext = "_Ew" if energy_weighted else ""

    plots_d: Dict[str, Figure] = {}
    for iftx, ftn in enumerate(conf.loader.x_features):
        if energy_weighted and iftx == epos:
            continue
        b = bins[iftx] if bins is not None else None
        simw = sim_features[ename] if energy_weighted else None
        genw = gen_features[ename] if energy_weighted else None
        plots_d[f"marginal_{ftn}{fext}.pdf"] = ratioplot(
            sim=sim_features[ftn],
            gen=gen_features[ftn],
            title=var_to_label(ftn),
            bins=b,
            simw=simw,
            genw=genw,
        )

    return plots_d


def _exftxt(arr):
    return {
        varname: arr
        for varname, arr in zip(
            conf.loader.x_features,
            arr.reshape(-1, conf.loader.n_features).T.cpu().numpy(),
        )
    }



import torch
from torch_scatter import scatter_mean

from fgsim.config import conf
from fgsim.utils import check_tensor

from .objcol import scaler


def rotate_alpha(alphas, batchidx, fake=False, center=False):
    batch_size = int(batchidx[-1] + 1)
    alphapos = conf.loader.x_features.index("alpha")
    ascalers = scaler.transfs_x[alphapos].steps[::-1]

    assert ascalers[0][0] == "standardscaler"
    mean = ascalers[0][1].mean_[0]
    scale = ascalers[0][1].scale_[0]

    # Backwards transform #0 stdscalar
    alphas = alphas.clone().double() * scale + mean

    # Backwards transform #1 logit
    alphas = torch.special.expit(alphas)

    assert (alphas <= 1).all()
    assert (0 <= alphas).all()
    # Rotation
    # smin, smax = ascalers[2][1].feature_range
    if not center:
        shift = torch.rand(batch_size).to(alphas.device)[batchidx]
    else:
        shift = -scatter_mean(alphas, batchidx)[batchidx] + 0.5
    if fake:
        shift *= 0
    alphas = alphas.clone() + shift
    alphas[alphas > 1] -= 1
    alphas[alphas < 0] += 1

    assert (alphas <= 1).all()
    assert (0 <= alphas).all()

    # Forward transform #1 logit
    alphas = torch.special.logit(alphas.clone())
    check_tensor(alphas)

    # Forward transform #0 stdscalar
    alphas = (alphas.clone() - mean) / scale
    check_tensor(alphas)
    return alphas.float()



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colors import LogNorm, Normalize
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1 import make_axes_locatable

from fgsim.plot.binborders import binborders_wo_outliers, chip_to_binborders
from fgsim.utils.torchtonp import wrap_torch_to_np

np.set_printoptions(formatter={"float_kind": "{:.3g}".format})


@wrap_torch_to_np
def hist2d(
    sim: np.ndarray,
    gen: np.ndarray,
    title: str,
    v1name: str,
    v2name: str,
    v1bins: Optional[np.ndarray] = None,
    v2bins: Optional[np.ndarray] = None,
    simw: Optional[np.ndarray] = None,
    genw: Optional[np.ndarray] = None,
    step: Optional[int] = None,
) -> Figure:
    plt.close("all")
    plt.cla()
    plt.clf()
    sns.set()
    fig: Figure
    sim_axes: Axes
    gen_axes: Axes
    fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)  # , figsize=(20, 12)
    (sim_axes, gen_axes) = axes
    if v1bins is None:
        xedges = binborders_wo_outliers(sim[:, 0])
        yedges = binborders_wo_outliers(sim[:, 1])
    else:
        xedges = v1bins
        yedges = v2bins

    # flip x and y axis if the x val has more bins
    if len(xedges) > len(yedges) + 5:
        xedges, yedges = yedges, xedges
        sim = sim[:, (1, 0)]
        gen = gen[:, (1, 0)]
        v1bins, v2bins = v2bins, v1bins
        v1name, v2name = v2name, v1name

    norm = None
    for iax, (ax, arr, weights) in enumerate(
        zip([sim_axes, gen_axes], [sim, gen], [simw, genw])
    ):
        mesh, norm = _2dhist_with_autonorm(
            ax=ax,
            x=chip_to_binborders(arr[:, 0], xedges),
            y=chip_to_binborders(arr[:, 1], yedges),
            bins=[xedges, yedges],
            cmap=plt.cm.hot,
            norm=norm,
            linewidths=2,
            weights=weights,
        )
        for spline in ax.spines.values():
            spline.set_linewidth(1.2)
            spline.set_color("black")

        if iax == 0:
            ax.set_ylabel(v2name)
        ax.set_xlabel(v1name)

    sim_axes.set_title("Simulation")
    gen_axes.set_title("Model")

    if step is not None:
        title += f"\nStep {step}"
    fig.suptitle(title)
    fig.tight_layout()

    cax = make_axes_locatable(plt.gca()).append_axes("right", "5%", pad="3%")

    fig.colorbar(mesh, cax)
    fig.tight_layout()
    # fig.savefig("wd/fig.pdf")
    return fig


def edge_to_labels(edges):
    if len(edges) < 10:
        ticks = np.arange(len(edges))
    else:
        ticks = np.linspace(0, len(edges) - 1, 10, dtype="int")
    labels = [str(int(e)) for e in edges[ticks]]
    return dict(ticks=ticks, labels=labels)


def _2dhist_with_autonorm(
    ax,
    x,
    y,
    bins,
    norm=None,
    range=None,
    density=False,
    weights=None,
    **kwargs,
):
    xedges, yedges = bins
    h, _, _ = np.histogram2d(
        x, y, bins=bins, range=range, density=density, weights=weights
    )

    if norm is None:
        if (h > (h.max() / 10)).mean() < 0.1:
            # if h.max() / max(np.median(h), 1) > 6:
            norm = LogNorm(1, h.max())
        else:
            norm = Normalize(0, h.max())

    pc = ax.pcolormesh(xedges, yedges, h.T, norm=norm, **kwargs)
    ax.set_xlim(xedges[0], xedges[-1])
    ax.set_ylim(yedges[0], yedges[-1])

    return pc, norm



import numpy as np

from fgsim.utils.torchtonp import wrap_torch_to_np


@wrap_torch_to_np
def binborders_wo_outliers(points: np.ndarray, bins=50) -> np.ndarray:
    assert len(points.shape) == 1
    if len(np.unique(points[:500])) < 50:
        uniques = np.sort(np.unique(points))
        if len(uniques) < 50:
            delta = (uniques[1] - uniques[0]) / 2
            return np.concatenate([uniques[0:1] - delta, uniques + delta])

    return np.linspace(*bounds_wo_outliers(points), num=bins, endpoint=True)


def bincenters(bins: np.ndarray) -> np.ndarray:
    return (bins[1:] + bins[:-1]) / 2


@wrap_torch_to_np
def bounds_wo_outliers(points: np.ndarray) -> tuple:
    median = np.median(points, axis=0)

    # med_abs_lfluk = np.sqrt(np.mean((points[points < median] - median) ** 2))
    # med_abs_ufluk = np.sqrt(np.mean((points[points > median] - median) ** 2))
    # upper = median + max(med_abs_ufluk,med_abs_ufluk)
    # lower = median - max(med_abs_ufluk,med_abs_ufluk)
    outlier_scale = (
        max(
            np.abs(np.quantile(points, 0.99) - median),
            np.abs(np.quantile(points, 0.01) - median),
        )
        * 1.1
    )
    upper = median + outlier_scale
    lower = median - outlier_scale
    # print(lower,np.min(points), upper,np.max(points))
    upper = np.min([upper, np.max(points)])
    lower = np.max([lower, np.min(points)])
    return lower, upper


def chip_to_binborders(arr, binborders):
    return np.clip(arr, binborders[0], binborders[-1])



from typing import List, Union

import queueflow as qf
from torch.multiprocessing import Queue, Value
from torch_geometric.data import Batch, Data

from fgsim.config import conf

from .objcol import contruct_graph_from_row, read_chunks, scaler

# Sharded switch for the postprocessing
shared_postprocess_switch = Value("i", 0)
shared_batch_size = Value("i", int(conf.loader.batch_size))


# Collect the steps
def process_seq() -> List[Union[qf.StepBase, Queue]]:
    return [
        qf.ProcessStep(read_chunks, 1, name="read_chunk"),
        qf.PoolStep(
            transform_and_scale,
            nworkers=conf.loader.n_workers_transform,
            name="transform",
        ),
        qf.RepackStep(shared_batch_size),
        qf.ProcessStep(aggregate_to_batch, 1, name="batch"),
        # Needed for outputs to stay in order.
        qf.ProcessStep(
            magic_do_nothing,
            1,
            name="magic_do_nothing",
        ),
        Queue(conf.loader.prefetch_batches),
    ]


# Methods used in the Sequence
# reading from the filesystem
def transform_and_scale(pc) -> Data:
    graph = contruct_graph_from_row(pc)
    graph.n_pointsv = (
        graph.y[..., conf.loader.y_features.index("num_particles")]
        .int()
        .reshape(-1)
    )
    graph.x = scaler.transform(graph.x, "x")
    graph.y = scaler.transform(graph.y, "y")

    return graph


def aggregate_to_batch(list_of_events) -> Batch:
    batch = Batch.from_data_list(list_of_events)
    return batch


def magic_do_nothing(batch: Batch) -> Batch:
    return batch



# %%
import numpy as np
import torch
from torch_geometric.data import Batch
from tqdm import tqdm

from fgsim.config import compute_conf, conf

compute_conf(conf, {"dataset_name": "calochallange", "command": "test"})
step = 1_000
conf.loader.scaling_fit_size //= 10
conf.loader.batch_size = step

from fgsim.io.dequantscaler import (  # noqa: E402
    FunctionTransformer,
    dequant,
    requant,
    scipy,
)
from fgsim.loaders.calochallange.objcol import (  # noqa: E402
    contruct_graph_from_row,
    file_manager,
    read_chunks,
    scaler,
)
from fgsim.loaders.calochallange.voxelize import (  # noqa: E402
    dims,
    sum_dublicate_hits,
    voxelize,
)

# scaler.fit(True)
# print("fit done")

for iftx in range(1, 4):
    sel_tf = scaler.transfs_x[iftx]
    sel_tf.steps[0][1].func = dequant
    sel_tf.steps[0][1].inverse_func = requant

    sel_tf.steps[2][1].func = scipy.special.logit
    sel_tf.steps[2][1].inverse_func = scipy.special.expit


fpos = np.arange(0, 100_000, step)
for i in tqdm(fpos):
    pcs = read_chunks([(file_manager.files[0], i, i + step)])
    gl = []
    for pc in pcs:
        graph = contruct_graph_from_row(pc)
        graph.n_pointsv = (
            graph.y[..., conf.loader.y_features.index("num_particles")]
            .int()
            .reshape(-1)
        )

        gl.append(graph)

    batch = Batch.from_data_list(gl)
    batch_untf = batch.clone()
    batch.y = scaler.transform(batch.y, "y")

    ## Manual scaling
    # maxscaler = 1
    maxscaler = 5
    x = batch.x.clone()
    # forward
    for iftx in range(4):
        ix = x[:, [iftx]].double().numpy()
        sel_tf = scaler.transfs_x[iftx]
        tfsteps = [e[1] for e in sel_tf.steps][: maxscaler + 1]

        for itf in tfsteps:
            x_tf = itf.transform(ix)
            x_tf_inv = itf.inverse_transform(x_tf)
            assert np.allclose(x_tf_inv, ix)
            ix = x_tf
        x[:, [iftx]] = torch.tensor(ix).float()

    x_forward = x.clone()

    # backward
    for iftx in range(4):
        ix = x[:, [iftx]].double().numpy()
        sel_tf = scaler.transfs_x[iftx]
        tfsteps = [e[1] for e in sel_tf.steps][: maxscaler + 1]

        for itf in tfsteps[::-1]:
            x_tf = itf.inverse_transform(ix)
            x_tf_inv = itf.transform(x_tf)

            if isinstance(itf, FunctionTransformer) and itf.func == dequant:
                assert np.allclose(x_tf_inv.astype("int"), ix.astype("int"))
            else:
                assert np.allclose(x_tf_inv, ix)
            ix = x_tf
        x[:, [iftx]] = torch.tensor(ix).float()

    assert torch.allclose(x, batch_untf.x)

    # batch.x = x
    batch.y = scaler.inverse_transform(batch.y, "y")

    ## Scaler scaling
    batch.x = scaler.transform(batch_untf.x.clone(), "x")
    batch_scaled = batch.clone()
    _inv = scaler.inverse_transform(batch.x, "x")
    assert torch.allclose(_inv, x)
    assert torch.allclose(_inv, batch_untf.x)
    batch.x = _inv

    # alphapos = conf.loader.x_features.index("alpha")
    # batch.x[..., alphapos] = rotate_alpha(
    #     batch.x[..., alphapos].clone(), batch.batch, True
    # )

    delta = (batch.x - batch_untf.x).abs()
    idx = torch.where(delta == delta.max())

    if not torch.allclose(batch.x, batch_untf.x, rtol=1e-04):
        sel_tf = scaler.transfs_x[idx[1]]
        print(batch.x[idx], batch_untf.x[idx], idx[1])
        print(
            batch_scaled.x[idx],
            sel_tf.transform(batch_untf.x[:, idx[1]]).squeeze()[idx[0]],
            idx[1],
        )

        tfsteps = [e[1] for e in sel_tf.steps]
        forward_res_step = [batch_untf.x[idx].numpy().reshape(-1, 1)]
        for _tf in tfsteps:
            forward_res_step.append(_tf.transform(forward_res_step[-1].copy()))

        backward_res_step = [batch_scaled.x[idx].numpy().reshape(-1, 1)]
        print(f"tranformed {backward_res_step[-1]}, scaled {forward_res_step[-1]}")
        for itf, _tf in enumerate(tfsteps[::-1]):
            backward_res_step.append(
                _tf.inverse_transform(backward_res_step[-1].copy())
            )
            a = backward_res_step[-1].squeeze()
            b = forward_res_step[-2 - itf].squeeze()
            print(a, b, "out of", _tf)

        raise Exception()

    assert torch.allclose(batch.x, batch_untf.x, rtol=1e-04)

    ## cc postprocess step by step

    # alphashift
    # alphas = batch.x[..., alphapos].clone()
    # shift = torch.randint(0, num_alpha, (batch.batch[-1] + 1,)).to(alphas.device)[
    #     batch.batch
    # ]
    # alphas = alphas.clone() + shift.float()
    # alphas[alphas > num_alpha - 1] -= num_alpha
    # batch.x[..., alphapos] = alphas

    batch = sum_dublicate_hits(batch, True)
    assert torch.allclose(batch.x, batch_untf.x)

    # batch = batch_to_Exyz(batch)
    # metrics: list[str] = conf.training.val.metrics
    # if "sphereratio" in metrics:
    #     batch["sphereratio"] = sphereratio(batch)
    # if "fpc" in metrics:
    #     batch["fpc"] = fpc_from_batch(batch)
    # if "showershape" in metrics:
    #     batch["showershape"] = analyze_layers(batch)
    # if "response" in metrics:
    #     batch["response"] = response(batch)

    assert torch.allclose(batch.x, batch_untf.x)
    pcs_out = voxelize(batch)
    input_images = torch.stack([e[1].reshape(*dims) for e in pcs], 0)
    assert torch.allclose(pcs_out.float(), input_images)



from typing import Dict, Optional, Union

import numpy as np

from fgsim.config import conf

labels_dict: Dict[str, str] = {
    "etarel": "$\\eta_\\mathrm{rel}$",
    "phirel": "$\\phi_\\mathrm{rel}$",
    "ptrel": "$p^\\mathrm{T}_\\mathrm{rel}$",
    "mass": "$m_{rel}$",
    "phi": "$Σ ϕ_{rel}$",
    "pt": "$Σp_{T}^{rel}$",
    "eta": "$Ση_{rel}$",
    "response": "Response (E/ΣE)",
    "showershape_peak_layer": "Peak Layer",
    "showershape_psr": "Ratio Δ(Turnon(/off), Peak) Layer",
    "showershape_turnon_layer": "Turnon Layer",
    "sphereratio": "Sphere(Δ=0.2)/Sphere(Δ=0.3)",
    "fpc_x": "First PCA vector x",
    "fpc_y": "First PCA vector y",
    "fpc_z": "First PCA vector z",
    "fpc_eval": "First PCA Eigenvalue",
}


def var_to_label(v: Union[str, int]) -> str:
    if isinstance(v, int):
        vname = conf.loader.x_features[v]
    else:
        vname = v
    if vname in labels_dict:
        return labels_dict[vname]
    else:
        return vname


def var_to_bins(v: Union[str, int]) -> Optional[np.ndarray]:
    if isinstance(v, int):
        vname = conf.loader.x_features[v]
    else:
        vname = v

    if (
        conf.dataset_name == "calochallange"
        and "calochallange2" in conf.loader.dataset_path
    ):
        return {
            "E": np.linspace(0, 6000, 100 + 1) - 0.5,  # E
            "z": np.linspace(0, 45, 45 + 1) - 0.5,  # z
            "alpha": np.linspace(0, 16, 16 + 1) - 0.5,  # alpha
            "r": np.linspace(0, 9, 9 + 1) - 0.5,  # r
        }[vname]
    return None



import torch
from torch_geometric.data import Batch
from torch_geometric.nn.pool import global_mean_pool

from fgsim.models.pool.std_pool import global_std_pool


def fpc_from_batch(batch: Batch) -> dict[str, torch.Tensor]:
    """Get the first principal component from a PC"""
    batchidx = batch.batch
    xyz = batch.xyz.double()
    means = global_mean_pool(xyz, batchidx)
    stds = global_std_pool(xyz, batchidx)
    deltas = (xyz - means[batchidx]) / (stds[batchidx] + 1e-8)
    cov = torch.stack(
        [
            torch.stack(
                [
                    global_mean_pool(deltas[:, i] * deltas[:, j], batchidx)
                    # global_add_pool(deltas[:, i] * deltas[:, j], batchidx)
                    # / (batch.ptr[1:]-batch.ptr[:-1] - 1)  # normalize
                    for i in range(3)
                ]
            )
            for j in range(3)
        ]
    ).transpose(0, -1)

    e_vals, e_vec = torch.linalg.eigh(cov)

    # largest_ev = e_val.argmax(-1).reshape(-1, 1, 1)
    # first_pc = e_vec.take_along_dim(largest_ev, -1)
    # (first_pc==e_vec[:,:,[2]]).all() why?? are they sorted?
    # https://pytorch.org/docs/stable/generated/torch.linalg.eigh.html
    # e_vec are sorted, get last colomn
    first_pc = e_vec[:, :, -1]

    # TEST
    # untrfs = deltas[batchidx==0]
    # first pc in the the first component:
    # fct = (first_pc[0].reshape(1, 3) @ untrfs.T).reshape(-1, 1) * first_pc[
    #     0
    # ].reshape(1, 3)
    # assert ((untrfs-fct).std(0)<(untrfs).std(0)).all()
    return dict(zip(["x", "y", "z"], first_pc.T.float().abs())) | {
        "eval": e_vals[:, -1].float()
    }



import torch
from torch_geometric.data import Batch
from torch_geometric.nn.pool import (
    global_add_pool,
    global_max_pool,
    global_mean_pool,
)
from torch_scatter import scatter_std

from fgsim.config import conf

n_layers = 45


def analyze_layers(batch: Batch) -> dict[str, torch.Tensor]:
    """Get the ratio between half-turnon and
    half turnoff to peak center from a PC"""
    batchidx = batch.batch
    n_events = int(batchidx[-1] + 1)
    device = batch.x.device
    z = batch.xyz[..., -1]

    # compute the histogramm with dim (events,layers)
    z_ev = (z.long() + batchidx * n_layers).sort().values

    zvals, zcounts = z_ev.unique_consecutive(return_counts=True)

    layer_idx = zvals % n_layers
    ev_idx = (zvals - layer_idx) // n_layers

    hist = torch.zeros(n_events, n_layers).long().to(device)
    hist[ev_idx, layer_idx] = zcounts

    assert (hist.sum(1) == batch.ptr[1:] - batch.ptr[:-1]).all()

    max_hits_per_event, peak_layer_per_event = hist.max(1)

    # produce an index for each layer in each event
    # centered_layers_perevent = (
    #     torch.arange(n_layers).to(device).reshape(1, -1).repeat(n_events, 1)
    # )
    # # shift peak to number of layers
    # centered_layers_perevent += -peak_layer_per_event.reshape(-1, 1) + n_layers
    # # separate idx by event
    # eventshift = (
    #     (torch.arange(n_events) * n_layers * 2)
    #     .to(device)
    #     .reshape(-1, 1)
    #     .repeat(1, n_layers)
    # )
    # centered_layers_perevent += eventshift

    # # aggegate for each of the shifted layers
    # zshape_centered = global_add_pool(
    #     hist.reshape(-1), centered_layers_perevent.reshape(-1)
    # )

    # # Estimate the turnon/turnoff width
    # get a boolean matrix to find the layers with enough hits
    occupied_layers = (
        hist > max_hits_per_event.reshape(-1, 1).repeat(1, n_layers) / 2
    )
    eventidx, occ_layer = occupied_layers.nonzero().T
    # find the min and max layer from the bool matrix ath the same time
    minmax = global_max_pool(torch.stack([occ_layer, -occ_layer]).T, eventidx)
    # reverse the -1 trick
    turnoff_layer, turnon_layer = (minmax * torch.tensor([1, -1]).to(device)).T

    # distance to the peak
    turnoff_dist = turnoff_layer - peak_layer_per_event
    turnon_dist = peak_layer_per_event - turnon_layer
    psr = (turnoff_dist + 1) / (turnon_dist + 1)

    return {
        "peak_layer": peak_layer_per_event.float(),
        "psr": psr.float(),
        "turnon_layer": turnon_layer.float(),
    }


def sphereratio(batch: Batch) -> dict[str, torch.Tensor]:
    batchidx = batch.batch
    Ehit = batch.x[:, conf.loader.x_ftx_energy_pos].reshape(-1, 1)

    e_small, e_large = __dist_fration(Ehit, batch.xyz, batchidx, 0.3, 0.8)

    return {
        "small": e_small,
        "large": e_large,
        "ratio": e_small / e_large,
    }


def cyratio(batch: Batch) -> dict[str, torch.Tensor]:
    batchidx = batch.batch
    Ehit = batch.x[:, conf.loader.x_ftx_energy_pos].reshape(-1, 1)

    e_small, e_large = __dist_fration(
        Ehit, batch.xyz[:, [0, 1]], batchidx, 0.2, 0.6
    )

    return {
        "small": e_small,
        "large": e_large,
        "ratio": e_small / e_large,
    }


def __dist_fration(Ehit, pos, batchidx, small, large, center_energy_weighted=True):
    Esum = global_add_pool(Ehit, batchidx).reshape(-1, 1)

    # get the center, weighted by energy
    if center_energy_weighted:
        center = global_add_pool(pos * Ehit, batchidx) / Esum
    else:
        center = global_mean_pool(pos, batchidx)
    std = scatter_std(pos, batchidx, dim=-2)
    # hit distance to center
    delta = (((pos - center[batchidx]) / std[batchidx]) ** 2).mean(-1).sqrt()
    del center, std
    # energy fraction inside circle around center
    e_small = (
        global_add_pool(Ehit.squeeze() * (delta < small).float(), batchidx)
        / Esum.squeeze()
    )
    e_large = (
        global_add_pool(Ehit.squeeze() * (delta < large).float(), batchidx)
        / Esum.squeeze()
    )
    # __plot_frac(e_small, e_large, small, large)
    return e_small, e_large


# def __plot_frac(e_small, e_large, small, large):
#     from matplotlib import pyplot as plt

#     ax: plt.Axes
#     fig, axes = plt.subplots(3, 1)
#     for ax, arr, title in zip(
#         axes,
#         [e_small, e_large, e_small / e_large],
#         [f"small {small}", f"large {large}", "ratio"],
#     ):
#         ax.hist(arr.cpu().numpy(), bins=100)
#         ax.set_title(title)
#     fig.tight_layout()
#     fig.savefig("wd/fig.pdf")
#     plt.close("all")


def response(batch: Batch) -> torch.Tensor:
    batchidx = batch.batch
    Ehit = batch.x[:, conf.loader.x_ftx_energy_pos]
    Esum = global_add_pool(Ehit, batchidx)

    return Esum / batch.y[:, conf.loader.y_features.index("E")]



from pathlib import Path
from typing import List, Tuple

import h5py
import numpy as np
import torch
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import (
    FunctionTransformer,
    PowerTransformer,
    QuantileTransformer,
    StandardScaler,
)
from torch_geometric.data import Data

from fgsim.config import conf
from fgsim.io import FileManager, ScalerBase
from fgsim.io.dequantscaler import dequant_stdscale


def path_to_len(fn: Path) -> int:
    return len(h5py.File(fn, "r")["incident_energies"])


file_manager = FileManager(
    path_to_len,
    files=list(
        Path(conf.loader.dataset_path).expanduser().glob(conf.loader.dataset_glob)
    ),
)


def readpath(
    fn: Path,
    start: int,
    end: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    with h5py.File(fn, "r") as electron_file:
        energies = electron_file["incident_energies"][start:end]
        showers = electron_file["showers"][start:end]
    res = (torch.Tensor(energies), torch.Tensor(showers))
    return res


def read_chunks(
    chunks: List[Tuple[Path, int, int]]
) -> List[Tuple[torch.Tensor, torch.Tensor]]:
    chunks_list = []
    for chunk in chunks:
        chunks_list.append(readpath(*chunk))
    res = (
        torch.concat([e[0] for e in chunks_list]),
        torch.concat([e[1] for e in chunks_list]),
    )
    return [(res[0][ievent], res[1][ievent]) for ievent in range(len(res[1]))]


def contruct_graph_from_row(chk: Tuple[torch.Tensor, torch.Tensor]) -> Data:
    E, shower = chk[0].clone(), chk[1].clone()
    num_z = 45
    num_alpha = 16
    num_r = 9
    shower = shower.reshape(num_z, num_alpha, num_r)
    idxs = torch.where(shower)
    h_energy = shower[idxs]
    z, alpha, r = idxs

    pc = torch.stack([h_energy, z, alpha, r]).T
    assert not pc.isnan().any()
    num_hits = torch.tensor(idxs[0].shape).float()
    res = Data(x=pc, y=torch.concat([E, num_hits, E / num_hits]).reshape(1, 3))
    return res


def Identity(x):
    return x


def LimitForBoxCox(x):
    return np.clip(x, -19, None)


hitE_tf = make_pipeline(
    PowerTransformer(method="box-cox", standardize=False),
    FunctionTransformer(Identity, LimitForBoxCox, validate=True),
    # SplineTransformer(),
    StandardScaler(),
)

E_tf = make_pipeline(
    PowerTransformer(method="box-cox", standardize=False),
    QuantileTransformer(output_distribution="normal"),
)
num_particles_tf = make_pipeline(
    *dequant_stdscale((0, conf.loader.n_points + 1)),
    QuantileTransformer(output_distribution="normal"),
)
E_per_hit_tf = make_pipeline(
    PowerTransformer(method="box-cox", standardize=False),
    QuantileTransformer(output_distribution="normal"),
)

scaler = ScalerBase(
    files=file_manager.files,
    len_dict=file_manager.file_len_dict,
    transfs_x=[
        hitE_tf,  # h_energy
        make_pipeline(*dequant_stdscale()),  # z
        make_pipeline(*dequant_stdscale()),  # alpha
        make_pipeline(*dequant_stdscale()),  # r
    ],
    transfs_y=[
        E_tf,  # Energy
        num_particles_tf,  # num_particles
        E_per_hit_tf,  # Energy pre hit
    ],
    read_chunk=read_chunks,
    transform_wo_scaling=contruct_graph_from_row,
)



from typing import Optional

import matplotlib.pyplot as plt
import mplhep
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from fgsim.utils.torchtonp import wrap_torch_to_np

from .binborders import binborders_wo_outliers, bincenters


@wrap_torch_to_np
def ratioplot(
    sim: np.ndarray,
    gen: np.ndarray,
    title: str,
    bins: Optional[np.ndarray] = None,
    simw: Optional[np.ndarray] = None,
    genw: Optional[np.ndarray] = None,
) -> Figure:
    if bins is None:
        bins = binborders_wo_outliers(sim)
    n_bins = len(bins) - 1

    sim_hist, _ = np.histogram(sim, bins=bins, weights=simw)
    gen_hist, _ = np.histogram(gen, bins=bins, weights=genw)
    sim_error = np.sqrt(sim_hist)
    gen_error = np.sqrt(gen_hist)

    plt.close("all")
    ax: Axes
    axrat: Axes
    fig, (ax, axrat) = plt.subplots(
        2, 1, figsize=(6, 7), gridspec_kw={"height_ratios": [2, 0.7]}, sharex="col"
    )

    mplhep.histplot(
        [sim_hist, gen_hist],
        bins=bins,
        label=["Simulation", "Model"],
        yerr=[sim_error, gen_error],
        ax=ax,
    )
    # overflow bins
    delta = (bins[1] - bins[0]) / 2
    simcolor = ax.containers[1][0]._color
    gencolor = ax.containers[2][0]._color
    kwstyle = dict(linestyle=(0, (0.5, 0.3)), lw=3)
    for arr, color, factor in zip([gen, sim], [gencolor, simcolor], [1, 2]):
        ax.vlines(
            x=bins[0] - factor * delta,
            ymin=0,
            ymax=(arr < bins[0]).sum(),
            color=color,
            **kwstyle,
        )
        ax.vlines(
            x=bins[-1] + factor * delta,
            ymin=0,
            ymax=(arr > bins[-1]).sum(),
            color=color,
            **kwstyle,
        )

    if (sim_hist > (sim_hist.max() / 10)).mean() < 0.1:
        ax.set_yscale("log")
    ax.set_ylabel("Frequency", fontsize=16)

    ax.legend(fontsize=16, loc="best")
    ax.tick_params(axis="both", which="major", labelsize=12)
    ax.tick_params(axis="both", which="minor", labelsize=10)

    # ratioplot
    with np.errstate(divide="ignore", invalid="ignore"):
        frac = gen_hist / sim_hist
        frac_error_y = np.abs(frac) * np.sqrt(
            (sim_error / sim_hist) ** 2 + (gen_error / gen_hist) ** 2
        )
        frac_error_x = np.array([(bins[1] - bins[0]) / 2.0] * n_bins)
        frac_mask = (frac != 0) & np.invert(np.isnan(frac_error_y))
    axrat.axhline(1, color="grey")

    axrat.errorbar(
        x=bincenters(bins)[frac_mask],
        y=frac[frac_mask],
        yerr=frac_error_y[frac_mask],
        xerr=frac_error_x[frac_mask],
        barsabove=True,
        linestyle="",
        marker=None,
        ecolor="black",
        markersize=2,
    )

    axrat.set_ylim(0.5, 1.5)
    axrat.xaxis.tick_top()
    for iax in [ax, axrat]:
        for spline in iax.spines.values():
            spline.set_linewidth(1)
            spline.set_color("black")
    if simw is not None:
        title += " weighted"
    fig.suptitle(title, fontsize=23)
    plt.tight_layout()
    # fig.savefig("wd/fig.pdf")
    return fig



import numpy as np
from scipy import sparse
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing._data import _handle_zeros_in_scale
from sklearn.utils.validation import FLOAT_DTYPES


class IdxToScale(MinMaxScaler):
    """Same as MinMaxScaler but fits min to the floor and Max to the roof"""

    # _parameter_constraints: dict = {
    #     "feature_range": [tuple],
    #     "copy": ["boolean"],
    #     "clip": ["boolean"],
    # }

    def __init__(self, feature_range=(0, 1), *, copy=True, clip=False):
        self.feature_range = feature_range
        self.copy = copy
        self.clip = clip

    def partial_fit(self, X, y=None):
        self._validate_params()

        feature_range = self.feature_range
        if feature_range[0] >= feature_range[1]:
            raise ValueError(
                "Minimum of desired feature range must be smaller than maximum."
                " Got %s."
                % str(feature_range)
            )

        if sparse.issparse(X):
            raise TypeError(
                "MinMaxScaler does not support sparse input. "
                "Consider using MaxAbsScaler instead."
            )

        first_pass = not hasattr(self, "n_samples_seen_")
        X = self._validate_data(
            X,
            reset=first_pass,
            dtype=FLOAT_DTYPES,
            force_all_finite="allow-nan",
        )

        # Change here
        data_min = np.floor(np.nanmin(X, axis=0))
        data_max = np.ceil(np.nanmax(X, axis=0))
        # Change end

        if first_pass:
            self.n_samples_seen_ = X.shape[0]
        else:
            data_min = np.minimum(self.data_min_, data_min)
            data_max = np.maximum(self.data_max_, data_max)
            self.n_samples_seen_ += X.shape[0]

        data_range = data_max - data_min
        self.scale_ = (
            feature_range[1] - feature_range[0]
        ) / _handle_zeros_in_scale(data_range, copy=True)
        self.min_ = feature_range[0] - data_min * self.scale_
        self.data_min_ = data_min
        self.data_max_ = data_max
        self.data_range_ = data_range
        return self



from functools import partial

import numpy as np
import scipy
from sklearn.preprocessing import FunctionTransformer, StandardScaler

from .idxscale import IdxToScale

__safety_gap = 1e-6


def dequant(x):
    assert x.dtype in [np.float64, np.int32]
    noise = np.random.rand(*x.shape)
    xnew = x.astype("float64") + np.clip(noise, __safety_gap, 1 - __safety_gap)
    assert np.all(x == np.floor(xnew))
    return xnew


def requant(x):
    assert x.dtype == np.float64
    # x_copy = x.copy()
    x = np.floor(x).astype("int32")
    # with the clip in dequant, the input must change
    # x[x == x_copy] -= 1
    # assert np.all(x.astype("int") == x)
    # delta = np.abs((x_copy - x))
    # assert (delta > 0).all() and (delta <= 1).all()
    return x


def forward(x, lower, dist):
    assert x.dtype == np.float64
    return (x - lower) / dist


def backward(x, lower, dist):
    assert x.dtype == np.float64
    return x * dist + lower


def idt(x):
    return x


def dequant_stdscale(inputrange=None) -> list:
    if inputrange is None:
        scaletf = IdxToScale((0, 1))
    else:
        lower, upper = inputrange
        dist = upper - lower

        scaletf = FunctionTransformer(
            partial(forward, lower=lower, dist=dist),
            partial(backward, lower=lower, dist=dist),
            check_inverse=True,
            validate=True,
        )
    tfseq = [
        FunctionTransformer(dequant, requant, check_inverse=True, validate=True),
        scaletf,
        FunctionTransformer(
            scipy.special.logit,  # scipy.special.erf,  # ,
            scipy.special.expit,  # scipy.special.erfinv,  #
            check_inverse=True,
            validate=True,
        ),
        StandardScaler(),
    ]
    return tfseq



from math import prod

import torch
from torch_scatter import scatter_add

from fgsim.config import conf, device
from fgsim.io.batch_tools import fix_slice_dict_nodeattr
from fgsim.utils.batch import ptr_from_batchidx

from .objcol import scaler

num_z = 45
num_alpha = 16
num_r = 9
dims = [num_z, num_alpha, num_r]

cell_idxs = torch.arange(prod(dims)).reshape(*dims)
full_event_cell_idx = (
    torch.arange(conf.loader.batch_size * dims[0] * dims[1] * dims[2])
    .reshape(conf.loader.batch_size, *dims)
    .to(device)
)


def voxelize(batch):
    batch_size = int(batch.batch[-1] + 1)
    x = batch.x

    # Get the valid hits
    # valid_hits = temp[~mask]
    # Ehit = valid_hits[:, 0]
    # valid_coordinates = valid_hits[:, 1:].long().t()
    # shower_index = torch.arange(batch_size).repeat_interleave(
    #     (~mask).float().sum(1).reshape(-1).int()
    # )
    shower_index = batch.batch
    Ehit = x.T[0]
    valid_coordinates = x.T[1:].int()
    indices = torch.cat((shower_index.unsqueeze(1), valid_coordinates.t()), dim=1)
    scatter_index = full_event_cell_idx[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    vox = scatter_add(
        src=Ehit,
        index=scatter_index,
        dim_size=prod(dims) * batch_size,
    )
    return vox.reshape(batch_size, *dims)


def invscale_add_position(batch):
    if batch.pos is not None:
        if len(batch.x) == len(batch.pos):
            return
    pos_l = []
    for iftx in range(batch.x.shape[1]):
        if iftx == conf.loader.x_ftx_energy_pos:
            continue
        pos_l.append(
            torch.tensor(
                scaler.transfs_x[iftx].inverse_transform(
                    batch.x[:, [iftx]].detach().cpu().double().numpy()
                )
            ).to(batch.x.device)
        )
    batch.pos = torch.hstack(pos_l).long()


def cell_occ_per_hit(batch):
    batch_size = int(batch.batch[-1] + 1)
    dev = batch.x.device
    fulldim = (batch_size, *dims)

    invscale_add_position(batch)
    indices = torch.hstack((batch.batch.unsqueeze(1), batch.pos))
    scatter_index = full_event_cell_idx[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    occ = scatter_add(
        src=torch.ones(len(batch.batch), dtype=torch.int, device=dev),
        index=scatter_index,
        dim_size=prod(dims) * batch_size,
    ).reshape(*fulldim)
    batch_occ = occ[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    return batch_occ


def sum_dublicate_hits(batch, forbid_dublicates=False):
    old_dev = batch.x.device
    batch = batch.to("cpu")
    dev = batch.x.device
    batchidx = batch.batch
    n_events = int(batchidx[-1] + 1)

    hitE = batch.x[:, 0]
    pos = batch.x[:, 1:].long()

    pos_z, pos_alpha, pos_r = pos.T
    assert (pos >= 0).all()
    assert (pos_z < num_z).all()
    assert (pos_alpha < num_alpha).all()
    assert (pos_r < num_r).all()

    cell_idx_per_hit = cell_idxs.to(dev)[pos_z, pos_alpha, pos_r]

    # give pos unique values for each event
    eventshift = torch.arange(n_events).to(dev) * prod(dims)
    event_and_cell_idxs = cell_idx_per_hit + eventshift[batchidx]

    # sort the event_and_cell_idxs
    cell_idx_per_hit, index_perm = _scatter_sort(event_and_cell_idxs, batchidx)
    hitE = hitE[index_perm]
    pos = pos[index_perm]
    assert (batchidx[index_perm] == batchidx).all()

    _, invidx, counts = torch.unique_consecutive(
        cell_idx_per_hit, return_inverse=True, return_counts=True
    )
    if forbid_dublicates:
        assert (counts - 1 == 0).all()

    hitE_new = scatter_add(hitE, invidx)
    sel_new_idx = counts.cumsum(-1) - 1
    if forbid_dublicates:
        assert (sel_new_idx == torch.arange(len(batch.x))).all()

    batchidx_new = batchidx[sel_new_idx]
    pos_new = pos[sel_new_idx]

    # count the cells, that have been hit multiple times
    n_multihit = scatter_add(counts - 1, batchidx_new)
    if forbid_dublicates:
        assert (n_multihit == 0).all()
    new_counts = torch.unique_consecutive(batchidx_new, return_counts=True)[1]

    x_new = torch.hstack([hitE_new.reshape(-1, 1), pos_new])

    # # Tests
    # old_counts = torch.unique_consecutive(batchidx, return_counts=True)[1]
    # if "n_pointsv" in batch.keys:
    #     assert (old_counts == batch.n_pointsv).all()
    # assert ((old_counts - new_counts) == n_multihit).all()
    # assert torch.allclose(
    #     scatter_add(hitE_new, batchidx_new), scatter_add(hitE, batchidx)
    # )
    # assert torch.allclose(
    #     scatter_add(pos_new * hitE_new.unsqueeze(-1), batchidx_new, -2),
    #     scatter_add(pos * hitE.unsqueeze(-1), batchidx, -2),
    # )
    if forbid_dublicates:
        assert (n_multihit == 0).all()
        assert (batch.batch == batchidx_new).all()
        assert (batch.n_pointsv == new_counts).all()
        for i in range(4):
            assert torch.allclose(batch.x.T[i][index_perm], x_new.T[i])

    batch.n_multihit = n_multihit
    batch.batch = batchidx_new
    batch.x = x_new
    batch.n_pointsv = new_counts
    # need to shift the ptr by the number of removed hits
    batch.ptr = ptr_from_batchidx(batchidx_new)

    batch.nhits = {
        "n": batch.n_pointsv,
        "n_by_E": batch.n_pointsv / batch.y[:, 0],
    }
    fix_slice_dict_nodeattr(batch, "x")

    return batch.to(old_dev)


def _scatter_sort(x, index, dim=-1):
    x, x_perm = torch.sort(x, dim=dim)
    index = index.take_along_dim(x_perm, dim=dim)
    index, index_perm = torch.sort(index, dim=dim, stable=True)
    x = x.take_along_dim(index_perm, dim=dim)
    return x, x_perm.take_along_dim(index_perm, dim=dim)


def test_sum_dublicate_hits():
    from torch_geometric.data import Batch, Data

    batch = Batch.from_data_list(
        [
            Data(
                x=torch.tensor(
                    [
                        [1, 0, 0, 0],
                        [1, 0, 1, 0],
                        [1, 0, 0, 0],
                        [1, num_z - 1, num_alpha - 1, num_r - 1],
                    ]
                ),
                y=torch.tensor([[1, 1]]),
                n_pointsv=torch.tensor(4),
            ),
            Data(
                x=torch.tensor(
                    [
                        [1, 0, 1, 0],
                        [1, 0, 0, 0],
                        [1, 0, 1, 0],
                    ]
                ),
                y=torch.tensor([[1, 1]]),
                n_pointsv=torch.tensor(3),
            ),
        ]
    )
    batch_new = sum_dublicate_hits(batch.clone())

    batchidx = batch.batch
    batchidx_new = batch_new.batch

    n_multihit = batch_new.n_multihit
    new_counts = batch_new.n_pointsv

    hitE = batch.x[:, 0]
    pos = batch.x[:, 1:].long()
    hitE_new = batch_new.x[:, 0]
    pos_new = batch_new.x[:, 1:].long()

    old_counts = torch.unique_consecutive(batchidx, return_counts=True)[1]
    if "n_pointsv" in batch.keys:
        assert (old_counts == batch.n_pointsv).all()
    assert ((old_counts - new_counts) == n_multihit).all()
    assert torch.allclose(
        scatter_add(hitE_new, batchidx_new), scatter_add(hitE, batchidx)
    )
    assert torch.allclose(
        scatter_add(pos_new * hitE_new.unsqueeze(-1), batchidx_new, -2),
        scatter_add(pos * hitE.unsqueeze(-1), batchidx, -2),
    )


test_sum_dublicate_hits()
